
go:-
	domain(Domain),

	output_counters_first_line,
	
	programs(Programs),
	member(Name,Programs),
	PathName = Name,
%	atom_concat('',Name,PathName),
	atom_concat(PathName,'.pl',File),
	format("\n------------ ~p -------------- \n",[File]),
	test_tabling(Domain,File), 
	test_plai(Domain,File),

	output_counters(Name),
	
	fail.

go.

test_tabling(Domain,File) :-
	abolish_all_tables,
	set_pp_flag(fixpoint,plai_tab),
	module(File),
	analyze(Domain).
test_plai(Domain,File) :-
	set_pp_flag(fixpoint,plai_wc),
	module(File),
	analyze(Domain).

:- use_module(library(stream_utils)).
:- use_module(library(between)).


num_samples(2).
num_discarded_samples(0).
	

main :-

	domain(Domain),

	output_nice_table_first_line,

	%% Make 10 iteraction for each program %%
	programs(Programs),
	member(Name,Programs),
	PathName = Name,
%	atom_concat('../tclp/',Name,PathName),
	atom_concat(PathName,'.pl',File),
	%%

	num_samples(NSamples),

	iterate_tab(NSamples,Domain,File,TabResults),
	iterate_plai(NSamples,Domain,File,PlaiResults),
	process_results(TabResults, Tab_Final_Results),
	process_results(PlaiResults, Plai_Final_Results),

%	output_raw_results(Name,Tab_Final_Results,Plai_Final_Results),
	output_nice_tables(Name,Tab_Final_Results,Plai_Final_Results),

	sum_column(Tab_Final_Results,Plai_Final_Results),
	
	fail.

main :-
	output_nice_table_last_line.

:- data column/2.

sum_column([Tavg|_],[Cavg|_]) :-
	column(T,C), !,
	retractall_fact(column(_,_)),
	NewT is Tavg + T,
	NewC is Cavg + C,
	asserta_fact(column(NewT,NewC)).
sum_column([Tavg|_],[Cavg|_]) :-
	asserta_fact(column(Tavg,Cavg)).
	




iterate_tab(0,_,_,[]).
iterate_tab(N,Domain,File,[T1|Rs]):-
	N > 0, N1 is N - 1,
	test_tabling_bak(Domain,File),
	plai_tab_time(T1), retractall_fact(plai_tab_time(T1)),
	iterate_tab(N1,Domain,File,Rs).

test_tabling_bak(Domain,File):- test_tabling(Domain, File), fail.
test_tabling_bak(_Domain,_File).

iterate_plai(0,_,_,[]).
iterate_plai(N,Domain,File,[T2|Rs]):-
	N > 0, N1 is N - 1,
	test_plai_bak(Domain,File),
	plai_time(T2), retractall_fact(plai_time(T2)),
	iterate_plai(N1,Domain, File,Rs).

test_plai_bak(Domain,File):- test_plai(Domain, File), fail.
test_plai_bak(_Domain,_File).





process_results(LR, [Avg,Mad|LR0]) :- !,
	revsorting(LR, LRSorted),
	num_discarded_samples(Discarded),
	length(DiscardedList, Discarded),
	append(DiscardedList, LR0, LRSorted),
%	remove_N_max(LR,LR0),
	average(LR0,Avg),
	mad(LR0,Mad).

revsorting([A|B], Sorted) :-
	revsorting(B, SortedTail),
	revinsert(A, SortedTail, Sorted).
revsorting([], []).

revinsert(A, [], [A]).
revinsert(A, [B|C], [B|D]) :-
	A @< B, !,
	revinsert(A, C, D).
revinsert(A, [B|C], [A,B|C]):- A @>= B.

% remove_max([_A],[]).
% remove_max([A,B|Ls],[A|Ls0]) :-
% 	A =< B,
% 	remove_max([B|Ls],Ls0).
% remove_max([A,B|Ls],[B|Ls0]) :-
% 	A > B,
% 	remove_max([A|Ls],Ls0).

average( List, Average ):- 
    sum( List, Sum ),
    length( List, Length ),
    Average is Sum / Length.

sum([X],X).
sum([X|Xs],S) :-
	sum(Xs,Ss),
	S is Ss + X.

mad(L,Mad) :-
	average(L,Avg),
	list_mad(L,Avg,Lmad),
	average(Lmad,Mad).

list_mad([],_,[]).
list_mad([X|Xs], Avg, [Mx|Mxs]):-
%	list_mad_(X,Avg, Mx),
	Mx is abs(X - Avg),
	list_mad(Xs,Avg,Mxs).
% list_mad_(X, Avg, Mx) :- 
% 	X =< Avg, !,
% 	Mx is Avg - X.
% list_mad_(X, Avg, Mx) :-
% 	X > Avg, !,
% 	Mx is X - Avg.

stdev(L, StdDev):-
	average(L, Avg),
	list_std(L, Avg, StdList),
	average(StdList, Variance),
	StdDev is sqrt(Variance).

list_std([],_,[]).
list_std([X|Xs], Avg, [Mx|Mxs]):-
	Mx is (X - Avg) ** 2,
	list_std(Xs,Avg,Mxs).


output_raw_results(File,Tab_Final_Results,Plai_Final_Results) :-
	raw_file(RawFile),
	open_output_file(Stream,RawFile,Current),
	format("\n------------ ~p -------------- \n",[File]),
	format("Version\t\tAvg\t\tMad\t\t1\t2\t3\t4\t5\t6\t7\t8\t9\n\n",[]),
	print_results('TABLING',Tab_Final_Results),
	print_results('CiaoPP ',Plai_Final_Results),
	close_output_file(Stream,Current).

print_results(Version,Results) :-	
	format("~p\t\t~2f\t\t~2f\t\t~2f\t~2f\t~2f\t~2f\t~2f\t~2f\t~2f\t~2f\t~2f\n\n",[Version|Results]).


output_nice_table_first_line :-
	nice_file(NiceFile),
	open_output_file(Stream,NiceFile,Current),
	format("\n\n\t~20|\tSpeedUp\t\tTABLING\t\t\tCiaoPP\n",[]),
	format("\t~20|\t\t\tms.\t(MAD)\t\tms.\t(MAD)\n",[]),
	close_output_file(Stream,Current).

output_nice_tables(File,[Tavg,Tmad|_],[Cavg,Cmad|_]) :-
	nice_file(NiceFile),
	open_output_file(Stream,NiceFile,Current),
	Speed is Cavg/Tavg,
	format("\t~p~20|\t~2f\t\t~2f\t(~2f)\t\t~2f\t(~2f)\n",[File,Speed,Tavg,Tmad,Cavg,Cmad]),
	close_output_file(Stream,Current).

output_nice_table_last_line :-
	nice_file(NiceFile),
	open_output_file(Stream,NiceFile,Current),
	column(TavgSum,CavgSum),
	Speed is CavgSum/TavgSum,
	format("\n\tAVG~20|\t~2f\t\t~2f\t      \t\t~2f\t      \n",[Speed,TavgSum,CavgSum]),
	close_output_file(Stream,Current).


output_counters_first_line :-
	counter_file(CounterFile),
	open_output_file(Stream,CounterFile,Current),
	format("\n\n \t ~20| CiaoPP ~40| TABLING\n",[]),
	format("\t ~20|# Calls~40|# Calls ~50|# Checks ~65|(# OKs) ~75|# AnsEnt ~85|# AnsJoin\n",[]),
	close_output_file(Stream,Current).

output_counters(File) :-
	counter_file(CounterFile),
	open_output_file(Stream,CounterFile,Current),
	(
	    plai_counter(inc_call,TotalCiaoPPCalls),
	    plai_tab_counter(inc_ans,TotalAnsEnt),
	    plai_tab_counter(inc_join,TotalAnsJoin),
	    plai_tab_counter(inc_call,TotalCalls),
	    plai_tab_counter(inc_check,TotalCheck),
	    plai_tab_counter(inc_ok,TotalOk) ->
	    format(" \t~p~20| ~2f~40| ~2f~50| ~2f ~65| ~2f ~75| ~2f ~85| ~2f\n",
	    [File, TotalCiaoPPCalls, TotalCalls, TotalCheck, TotalOk, TotalAnsEnt, TotalAnsJoin])
	;
	    true
	),
	close_output_file(Stream,Current).

open_output_file(Stream,File,Current) :-
	current_output(Current),
	open(File,append,Stream),
	set_output(Stream).
close_output_file(Stream,Current) :-
	set_output(Current),
	close(Stream).


