/* from the Gabriel Benchmark */  
/* obtained from Tep Dobry */
/* modified by Herve' Touati 01/15/87 */

:- module(browse,[investigate/2],[assertions]).

%% :- entry investigate(X,
%% 		[[star(SA),B,star(SB),B,a,star(SA),a,star(SB),star(SA)],
%% 		 [star(SA),star(SB),star(SB),star(SA),[star(SA)],[star(SB)]],
%% 		 [_X,_Y,star(_Z),[b,a],star(_A),_B,_C]
%% 		]) : ground(X).

:- entry investigate/2 : ground * list.

investigate([],_).
investigate([U|Units],Patterns) :-
	property(U,pattern,Data),
	p_investigate(Data,Patterns),
	investigate(Units,Patterns).

get_pats(Npats,Ipats,Result) :- 
	get_pats_(Npats,Ipats,Result,Ipats).

get_pats_(0,_X,[],_Y) :- !.
get_pats_(N,[X|Xs],[X|Ys],Ipats) :-
	N1 is N - 1,
	get_pats_(N1,Xs,Ys,Ipats).
get_pats_(N,[],Ys,Ipats) :-
	get_pats_(N,Ipats,Ys,Ipats).

property([],_X,_Y) :- fail.  /* don't really need this */
property([Prop|_RProps],P,Val) :-
	functor(Prop,P,_X),!,
	arg(1,Prop,Val).
property([_X|RProps],P,Val) :-
	property(RProps,P,Val).

p_investigate([],_X).
p_investigate([D|Data],Patterns) :-
	p_match(Patterns,D),
	p_investigate(Data,Patterns).

p_match([],_).
p_match([P|Patterns],D) :-
	(match(D,P), fail; true),
	p_match(Patterns, D).

match([],[]) :- !.
match([X|PRest],[Y|SRest]) :-
	var(Y),!,X = Y,
	match(PRest,SRest).
match(List,[Y|Rest]) :- 
	nonvar(Y),Y = star(X),!,
	concat(X,SRest,List),
	match(SRest,Rest).
match([X|PRest],[Y|SRest]) :-
	(atom(X) -> X = Y; match(X,Y)),
	match(PRest,SRest).

concat([],L,L).
concat([X|L1],L2,[X|L3]) :- concat(L1,L2,L3).

% ----------------------------------------------------------------------
